import threading


def set_timeout(timeout: float, function: callable, args: list = None,
                kwargs: dict = None) -> threading.Timer:
    """
    Wrapper for library thread timer. Easy to find for future reference and
    sets timer to be a daemon so that it will stop if the main thead
    terminates.

    NB: The interval the timer will wait before executing its action may not
    be exactly the same as the interval specified by the user.

    :param timeout: Number of seconds to delay before running the timer
    :param function: The function to be called
    :param args: The positional arguments for the function
    :param kwargs: The key word arguments for the function
    :return: The timer object so that calling cancel on it will stop the timer
    """
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    timer = threading.Timer(timeout, function, args, kwargs)

    # So it won't prevent program termination because it hasn't run
    timer.daemon = True

    timer.start()
    return timer


def set_interval(interval: float, function: callable, args: list = None,
                 kwargs: dict = None) -> threading.Event:
    """
    Repeated timer every interval seconds.

    NB: The interval the timer will wait before executing its action may not
    be exactly the same as the interval specified by the user.

    :param interval: Number of seconds to delay before running the timer
        each time
    :param function: The function to be called
    :param args: The positional arguments for the function
    :param kwargs: The key word arguments for the function
    :return: A thread event able to cancel the timer (e.g. result.set())
    """
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    class MyThread(threading.Thread):
        def __init__(self, event):
            threading.Thread.__init__(self)
            self.stopped = event

        def run(self):
            while not self.stopped.wait(interval):
                function(*args, **kwargs)  # call a function

    stop_event = threading.Event()
    t = MyThread(stop_event)
    t.daemon = True
    t.start()
    return stop_event



import logging
import os
import sys
from time import sleep

from opylib.log import log

if sys.platform.startswith('linux'):
    try:
        # noinspection PyUnresolvedReferences
        from google.colab import output


        def make_sound(_):
            output.eval_js(
                'new Audio('
                '"https://upload.wikimedia.org/wikipedia/commons/0/05'
                '/Beep-09.ogg").play()')


        def notify_err(_):
            make_sound()

    except ModuleNotFoundError:
        def make_sound(sleep_time=3):
            os.system('spd-say "Process finished"')
            sleep(sleep_time)


        def notify_err(sleep_time=0):
            os.system('spd-say "Errors detected"')
            sleep(sleep_time)

elif sys.platform.startswith('darwin'):
    def make_sound(sleep_time=3):
        os.system('say "Process finished"')
        sleep(sleep_time)


    def notify_err(sleep_time=0):
        os.system('say "Errors detected"')
        sleep(sleep_time)

elif sys.platform.startswith('win') or sys.platform.startswith('cygwin'):
    import winsound


    def make_sound(_=None):
        for i in range(1, 5):
            winsound.Beep(i * 400, 200)
        for i in range(5, 1, -1):
            winsound.Beep(i * 400, 200)


    def notify_err(_=None):
        for i in range(2):
            winsound.Beep(2000, 500)
            winsound.Beep(400, 200)
else:
    def make_sound(_=None):
        log(Exception(f'No Sound because unexpected OS: {sys.platform}'),
            logging.WARN)


    def notify_err(_=None):
        make_sound()



import logging

from opylib.log import log
from opylib.timer import set_timeout


class RateLimitedExecution:
    """
    Designed to run a functions with a maximum frequency.
    Will only run the first request per function until that function has
    been run. Uses a singleton so that the client classes do not need to
    keep a reference to the instance because of challenges with serialising
    classes that have references to threads. (Error msg: "TypeError:
    cannot pickle '_thread.lock' object")
    """
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = RateLimitedExecution()
        return cls._instance

    def __init__(self):
        self.pending = {}

    def _execute(self, function: callable, args: list, kwargs: dict):
        self.pending.pop(function)
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        function(*args, **kwargs)

    def register(self, timeout: float, function: callable, args: list = None,
                 kwargs: dict = None) -> bool:
        """
        Registers function if it is not currently pending
        :param timeout: Number of seconds to delay before running the timer
        :param function: The function to be called
        :param args: The positional arguments for the function
        :param kwargs: The key word arguments for the function
        :return: True if was not already pending otherwise False
        """
        if self.is_pending(function):
            log(f'[RateLimit] Request to register ignored. Already pending '
                f'{function}', logging.DEBUG)
            return False
        else:
            self.pending[function] = \
                set_timeout(timeout, self._execute, [function, args, kwargs])
            log(f'[RateLimit] Registered {function} with {timeout} second '
                f'timeout', logging.DEBUG)
            return True

    def is_pending(self, function: callable):
        return function in self.pending

    def cancel(self, function: callable) -> bool:
        """
        Attempts to cancel the function call.
        :param function: The function to cancel the call to
        :return: True if the function was pending else False
        """
        if self.is_pending(function):
            self.pending[function].cancel()
            self.pending.pop(function)
            log(f'[RateLimit] Canceled {function}', logging.DEBUG)
            return True
        else:
            log(f'[RateLimit] Not Pending. Did NOT cancel {function}',
                logging.DEBUG)
            return False



from opylib.dict_persistent import DictPersistent
from opylib.log import log


def get_db() -> dict:
    """
    Access to REPL DB
    :return:
    """
    try:
        # noinspection PyUnresolvedReferences
        from replit import db
        log("Imported access to REPL DB")
    except ModuleNotFoundError:
        db = DictPersistent()
        log("Unable to get REPL DB Using Local dict")
    # noinspection PyTypeChecker
    return db



import random


def shuffle(*iterables):
    """
    Takes one or more iterable objects and shuffles them keeping their items
    that were at the same index as another item still at the same index as
    the other item.

    WARNING: Uses zip and will reduce the length of the items to the length
    of the shortest among them

    :iterables: The items to be shuffled
    :return: the pair wise same but shuffled iterables
    """
    join = list(zip(*iterables))
    random.shuffle(join)
    return zip(*join)



# Most of the dict merging code was taken or based on
# https://stackoverflow.com/questions/38987/how-do-i-merge-two-dictionaries
# -in-a-single-expression-taking-union-of-dictiona


def merge_dicts(*dict_args):
    """
    This function is not meant to be used but rather to serve as a reminder to
    use the syntax {**x, **y, **z}.  This function works but is not as fast as
    that syntax by as much as twice.

    Given any number of dictionaries, shallow copy and merge into a new dict,
    precedence is given to key-value pairs in latter dictionaries.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def merge_two_dicts(d1, d2):
    """
    Slower implementation that works on older versions of python
    :param d1: Dict 1
    :param d2: Dict 2
    :return: Copy of d1 with new keys from d2 and overlap overwritten by d2
    """
    result = d1.copy()
    result.update(d2)
    return result


def merge_dict2(d1: dict, d2: dict) -> dict:
    """
    Merges exactly 2 dicts (For use when the syntax can't be remembered to
    do it inline as {**d1, **d2}
    NB: If the same key exists in both then d2 takes precedence
    :param d1: Dict 1
    :param d2: Dict 2
    :return: {**d1, **d2}
    """
    return {**d1, **d2}



import ntpath
import os
from pathlib import Path


def mkdir(folder):
    Path(folder).mkdir(parents=True, exist_ok=True)


def change_filename_ext(fn: str, new_ext: str) -> str:
    """
    Returns the filename with the new extension
    :param fn: The filename with the old extension
    :param new_ext: The new extension to put on the filename
    :return: The filename with the new extension
    """
    result, _ = os.path.splitext(fn)
    return f'{result}{new_ext}'


def ensure_parent_folder_exists(fn: str):
    """
    Expects fn to be a file name. Removes the file name and ensures that
    it's parent folder exists
    :param fn: Filename to use to find parent folder
    :return:
    """
    folder, _ = ntpath.split(fn)
    mkdir(folder)


def ensure_output_folder_empty(folder):
    """
    Ensure that output exists and does not already contain files
    Usually used to ensure that output is not overwritten
    NB: If output folder does not already exist. It is created.
    :param folder:
    :return:
    """
    if os.path.exists(folder):
        listing = os.listdir(folder)
        if len(listing) > 0:
            raise Exception(
                f'Output Folder ("{folder}") is '
                f'not empty: {len(listing)} file(s) found')
    else:
        mkdir(folder)



from typing import List


def get_attr_multi_level(obj, name: str):
    """
    Returns the attribute 'name' from obj. Support attributes of attributes
    separated by '.' So if you have a top level class MyClass which has a
    subclass Sub1 and that one has a subclass Sub2 and that has an attribute
    address you could access it with
    get_attr_multi_level(MyClass, 'Sub1.Sub2.address')
    :param obj: The base object to get name from
    :param name: The name of desired attribute (can be multiple levels deep)
    :return: The desired attribute
    """
    names = name.split('.')
    for x in names:
        obj = getattr(obj, x)
    return obj


def str_to_class(cls_name: str) -> type:
    """
    Converts a string into the class that it represents

    NB: Code based on https://stackoverflow.com/questions/452969/does-python
    -have-an-equivalent-to-java-class-forname
    :param cls_name: The string representation of the desired class
    :return: A pointer to the class (Able to be used as a constructor)
    """
    parts = cls_name.split('.')
    modules = '.'.join(parts[:-1])
    result = __import__(modules)
    for comp in parts[1:]:
        result = getattr(result, comp)
    return result


def strs_to_classes(cls_names: List[str]) -> List[type]:
    result = []
    for s in cls_names:
        result.append(str_to_class(s))
    return result


def dsv_line_to_list(line: str, *, delimiter=',', quote='"') -> List[str]:
    """
    Splits line into fields on delimiter ignoring delimiters in fields that
    start and end with quote

    NB: Empty fields produce an empty string

    :param line: The line to be split
    :param delimiter: The delimiter to use to split the fields
    :param quote: The quote char to surround fields that contain the delimiter
    :return: a list of the fields found
    """
    result = []
    within_quoted_field = False
    at_start_of_field = True
    last_was_quote = False  # Used to see if quote is not at end of field
    field = ''

    def new_field():
        nonlocal field, within_quoted_field, at_start_of_field, last_was_quote
        result.append(field)
        within_quoted_field = False
        at_start_of_field = True
        last_was_quote = False
        field = ''

    for char in line:
        if at_start_of_field:
            at_start_of_field = False
            # Check for quote
            if char == quote:
                within_quoted_field = True
                continue  # Skip quote do not include in field

        if within_quoted_field:
            if char == quote:
                last_was_quote = True
                continue  # May not want to add this char if end of field
            if last_was_quote:
                if char == delimiter:
                    new_field()
                    continue
                else:
                    field += quote
                    last_was_quote = False
            field += char
        else:
            if char == delimiter:
                new_field()
            else:
                field += char

    # Add last field that was being filled (or empty if empty in comma)
    result.append(field)

    return result



from datetime import datetime, time, timezone
from typing import Union


def get_now_str(fmt: str = "%Y-%m-%d %H%M") -> str:
    return datetime.now().strftime(fmt)


def is_aware(value: Union[datetime, time]) -> bool:
    """
    Returns true if an object is timezone aware else false

    Throws an exception if an object of unexpected type is received

    Implemented based on information found in official documentation at
    https://docs.python.org/3/library/datetime.html in the section
    "Determining if an Object is Aware or Naive"
    :param value: The object to determine if it is timezone aware
    :return: true if an object is timezone aware else false
    """
    if isinstance(value, datetime):
        return value.tzinfo is not None \
               and value.tzinfo.utcoffset(value) is not None
    if isinstance(value, time):
        return value.tzinfo is not None \
               and value.tzinfo.utcoffset(None) is not None
    raise TypeError(
        f'Expected object of type datetime or time but got {value.__class__}')


def make_aware(
        value: Union[datetime, time],
        def_timezone: timezone = timezone.utc) -> \
        Union[datetime, time]:
    """
    Makes the parameter passed timezone aware if it wasn't already
    :param value: The object to be made timezone aware
    :param def_timezone: The timezone to set if the object is not yet
    timezone aware
    :return: An object that is timezone aware
    """
    if not is_aware(value):
        value = value.replace(tzinfo=def_timezone)
        assert is_aware(value)
    return value



from typing import Callable

from opylib.log import log, log_exception, setup_log
from opylib.notify import make_sound
from opylib.stopwatch import StopWatch


def main_runner(main: Callable, *, should_complete_notify: bool = True):
    """
    Takes care of setup, exception and wind down boilerplate code.
    Example:

    if __name__ == '__main__':
        main_runner(main)

    :param main: The function to be called that will actually do the work
    :param should_complete_notify: Controls if a completed notification is given
    :return:
    """
    setup_log()
    sw = StopWatch('MAIN')
    try:
        main()
    except Exception as e:
        log_exception(e)
    sw.end()
    log('Run COMPLETE.')
    if should_complete_notify:
        make_sound()



# Couldn't use shelve because close is never called in client code as it is 
# meant to be transparent
import logging
from os import path

from opylib.data_import_export import load_yaml, save_yaml


class DictPersistent:
    """
    Simple dict that reads from a file on object creation and writes to the
    file each time a value is set.

    ASSUMPTION: Mutated objects do not need to be written to disk until a new
    request to set is received

    NB: Reads are cheap because it is kept in memory but every set writes
    everything to disk, so set is expensive.
    """

    def __init__(self, file_name: str = 'dict_persistent.yaml'):
        self.filename = file_name
        if not path.exists(file_name):
            with open(file_name, "x"):
                pass  # create the file (fail early if unable to create)
            self.data = {}
        else:
            self.data = load_yaml(file_name)
            if not isinstance(self.data, dict):
                logging.log(logging.ERROR,
                            f'Data loaded from {file_name} was not a dict. '
                            f'Reverting to empty dict')
                self.data = {}

    def __contains__(self, item):
        return item in self.data

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value
        self.save()

    def get(self, key):
        return self.data.get(key)

    def keys(self):
        return self.data.keys()

    def pop(self, key):
        result = self.data.pop(key)
        self.save()
        return result

    def save(self):
        save_yaml(self.data, self.filename)



import logging
from datetime import datetime

import yaml

from opylib.log import log
from opylib.timer import set_timeout

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


class DBCache:
    """
    Provides a write cache for dict like databases.
    ASSUMPTIONS:
        - Keys are not tuples (tuples are used to trigger conversion to yaml
            (see __setitem__ for more info)
        - Database is not updated elsewhere
        - Thead safety is not an issue, uses blocking routines expected to
            complete in a single pass (operations are not atomic)
    """

    def __init__(self, db_backing: dict, save_cache_delay: int = 30):
        self.save_cache_delay = save_cache_delay
        self.db_backing = db_backing
        self.cache = {}  # Values sent but not written to db yet

        # values that need to be converted before submitting to db
        self.to_yaml = set()

        self._timer_write_out = None
        self.last_write_time = datetime.now()

    def __contains__(self, item):
        return item in self.cache or item in self.db_backing

    def __setitem__(self, key, value):
        """
        Sets the value for key to the value passed. NB: If yaml is needed then
            a tuple should be passed as the key. The first element of the tuple
            should be the key and the second the value of should_yaml EG. db[
            'key', True] = 'apple'
        :param key: The key to set or tuple of form (key, should_yaml)
        :param value: The value to set
        """
        log(f'[DB Cache] received request to set {key}', logging.DEBUG)
        if isinstance(key, tuple):
            key, should_yaml = key
        else:
            should_yaml = False
        self.cache[key] = value
        if should_yaml:
            self.to_yaml.add(key)

        if not self.is_write_pending:
            sec_before_save_allowed = \
                self.save_cache_delay \
                - (datetime.now() - self.last_write_time).total_seconds()
            if sec_before_save_allowed < 0:
                self._write_to_backing()
            else:
                self._timer_write_out = set_timeout(sec_before_save_allowed,
                                                    self._write_to_backing)

    def get(self, key, should_yaml=False):
        """
        Retrieves a value corresponding to the key passed if not present
            returns None.
        :param key: The key to use to retrieve the value
        :param should_yaml: if the value need to be converted from yaml
        :return: The value corresponding to the key passed
        """
        if key in self.cache:
            result = self.cache[key]
        else:
            result = self.db_backing.get(key)

        if should_yaml and result is not None:
            return yaml.load(result, Loader=Loader)
        else:
            return result

    def __getitem__(self, key):
        """
        Gets the value for they key specified. NB: If yaml is needed then
            a tuple should be passed as the key. The first element
            of the tuple should be the key and the second the value of
            should_yaml EG. db['key', True] = 'apple'
        :param key: The key to set or tuple of form (key, should_yaml)
        :return:
        """
        if isinstance(key, tuple):
            key, should_yaml = key
        else:
            should_yaml = False

        if key in self.cache:
            # No concern for yaml because not written to db yet
            return self.cache[key]
        else:
            if key in self.db_backing:
                # to make use of yaml code there
                return self.get(key, should_yaml=should_yaml)
            else:
                return self.db_backing[key]  # To trigger correct exception

    def _write_to_backing(self):
        """
        Writes the cache to the db. Using existence in self.to_yaml to know
            if that value has to converted to a yaml string before saving to db
        ASSUMPTION: Function completes in its entirety without any other
            thread calling any of the other functions
        """
        log('[DB Cache] Purging', logging.DEBUG)
        for key in self.cache.keys():
            value = self.cache[key]
            if key in self.to_yaml:
                value = yaml.dump(value, Dumper=Dumper)
            self.db_backing[key] = value

        # Register save time and clear timer variable
        self.last_save_time = datetime.now()
        self._timer_write_out = None

        # reset cache and to_yaml
        self.cache = {}
        self.to_yaml = set()

    @property
    def is_write_pending(self):
        return self._timer_write_out is not None

    def keys(self):
        if self.is_write_pending:
            self._timer_write_out.cancel()
            self._write_to_backing()
        return self.db_backing.keys()



from datetime import datetime

from opylib.log import log


class StopWatch:
    def __init__(self, name, *, should_log_end=False, start_log_level=None):
        self.start_time = datetime.now()
        self.name = name
        self.end_time = None
        self.should_log_end = should_log_end
        self.run_msg = "Still Running"
        log(f"{self.name} stop watch start time: {self.start_time}",
            start_log_level)

    def end(self):
        self.end_time = datetime.now()
        self.run_msg = f"{self.name} stop watch runtime was {self.runtime()}"
        if self.should_log_end:
            log(f"{self.name} stop watch end time: {self.end_time}")
        log(self.run_msg)

    def runtime(self):
        return self.end_time - self.start_time

    def as_float(self):
        return self.runtime().total_seconds()



import logging
import re
from typing import Any, List, Type

import yaml

from opylib.files_folders import ensure_parent_folder_exists

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


def get_matlab_safe_list(org_list: List) -> List:
    """
    Take a list and returns a new list that is matlab safe
    :param org_list: input list to be made safe
    :return: list that is afe to save in matlab
    """
    result = []
    for val in org_list:
        if val is None:
            val = str(val)
        elif isinstance(val, dict):
            val = get_matlab_safe_dict(val)
        result.append(val)
    return result


def get_matlab_safe_dict(org_dict: dict):
    """
    Takes a dict and returns a new dict that is matlab safe.
    Removes None and validates keys.

    NB: keys may collide if the resolve to the same "clean" string. Warning
    logged but conversion not aborted
    :param org_dict: input dict to be made safe
    :return: dict that is safe to save in matlab
    """
    result = {}
    for key in org_dict.keys():
        valid_key = get_valid_ident(key)
        value = org_dict[key]
        if value is None:
            value = str(value)
        elif isinstance(value, dict):
            value = get_matlab_safe_dict(value)
        elif isinstance(value, list):
            value = get_matlab_safe_list(value)

        if result.get(valid_key) is not None:
            logging.warning(
                f'Got DUPLICATE key in dict with str rep of "{valid_key}"')
        result[valid_key] = value

    return result


def get_valid_ident(org):
    """Turns the parameter passed into a valid identifier"""
    result = re.sub(r' ', '_', str(org))
    result = re.sub(r'[^A-Za-z0-9_]', '', result)
    return result


def public_members_as_dict(class_: Type[Any]) -> dict:
    """
    Converts a class into a dict of its public values
    :param class_: the class to be converted
    :return: the public values from the class
    """
    result = {}
    for i in class_.__dict__.items():
        if not (i[0][0] == "_" or isinstance(i[1], classmethod)):
            if not isinstance(i[1], type):
                result[i[0]] = i[1]
            else:
                result[i[0]] = public_members_as_dict(i[1])
    return result


def save_list_to_file(file_list: List[str], file_name: str, end: str = ''):
    """
    Saves file_list to a file with file_name and appends end
    :param file_list: List of string to be saved to the file
    :param file_name: Name of file to be created or overwritten
    :param end: Optional line ending to append to each line
    :return:
    """
    ensure_parent_folder_exists(file_name)
    with open(file_name, 'w') as f:
        for item in file_list:
            f.write('%s%s' % (item, end))


def save_yaml(data, fn: str):
    """
    Saves data to yaml file named fn. Overwrites if it already exists
    :param data: Data to be saved
    :param fn: The name of the file to create
    """
    with open(fn, 'w') as f:
        f.write(yaml.dump(data, Dumper=Dumper))


def load_yaml(fn: str):
    """
    Loads data from a yaml file named fn
    :param fn: The name of the file to load from
    :return: The data from the file
    """
    with open(fn, 'r') as f:
        return yaml.load(f, Loader=Loader)



"""
Collection of functions that work on streaming data (data we only see once
one element at a time and cannot store for processing or regular processing
may overflow type limits)

NB: Some functions may only be able to give an approximation.
"""


def mean(curr_avg: float, n: int, new_value: float) -> float:
    """
    Updates an average value using the real time streaming algorithm which
    uses constant space for calculations and doesn't overflow a sum counter.
    Gives the exact value not an approximation.

    NB: Value of curr_avg doesn't matter for first call as long as n is set
    to 0 for first call

    :param curr_avg: The current average value
    :param n: The number of values that have been summed so far
    :param new_value: The value to be added to the average
    :return: The newly updated average value
    """
    return (curr_avg * n + new_value) / (n + 1)



import atexit
import logging
import os.path
import sys
from datetime import datetime

from opylib.files_folders import mkdir

_logger = logging.getLogger()

# Store directory where log was created.
# Needs to be converted to absolute path in setup in case working directory
# changes during execution. Used later for location to create file for errors
_log_dir: str = 'log/'


def log(msg, log_level=None):
    global _logger
    if log_level is None:
        log_level = logging.INFO
    if log_level >= logging.ERROR:
        _ErrorsStats.register_error_occurrence()
    _logger.log(log_level, msg)


def log_exception(err: Exception):
    global _logger
    _ErrorsStats.register_error_occurrence()
    _logger.exception(err)


def setup_log(filename=None, *, only_std_out=False,
              fmt_std_out='%(asctime)s %(levelname)s: %(message)s',
              fmt_file='%(asctime)s %(levelname)s: %(message)s',
              fmt_err='%(asctime)s %(levelname)s: %(message)s',
              error_notification_callback: callable = None):
    """
    Setups up logging handlers. Only needs to be called once

    :param filename: The file to save logs to
    :param only_std_out: If True does not log to file
    :param fmt_std_out: Format to use for Standard Out
    :param fmt_file: Format to use for file logging
    :param fmt_err: Format to use for Errors
    :param error_notification_callback: Called at the end of execution if
    there were errors
    :return:
    """
    global _logger, _log_dir

    if _logger.hasHandlers():
        log(f'{"!" * 20} REUSING LOG {"!" * 92}')
        return

    if filename is None:
        filename = \
            f'{_log_dir}run {datetime.now().strftime("%Y-%m-%d %H-%M-%S")}.log'
    _log_dir = os.path.abspath(os.path.dirname(filename))

    if not only_std_out:
        mkdir(_log_dir)
        # Set up a file handler
        file_handler = logging.FileHandler(filename)
        file_handler.setFormatter(
            logging.Formatter(fmt_file))
        file_handler.setLevel(logging.DEBUG)
        _logger.addHandler(file_handler)

    # Set up standard output
    std_stream_handler = logging.StreamHandler(sys.stdout)
    std_stream_handler.setFormatter(
        logging.Formatter(fmt_std_out))
    std_stream_handler.setLevel(logging.DEBUG)
    _logger.addHandler(std_stream_handler)

    # Set up error output
    err_stream_handler = logging.StreamHandler(sys.stderr)
    err_stream_handler.setFormatter(
        logging.Formatter(fmt_err))
    err_stream_handler.setLevel(logging.ERROR)
    _logger.addHandler(err_stream_handler)

    atexit.register(_check_error)

    if error_notification_callback is not None:
        _ErrorsStats.register_notification_callback(error_notification_callback)

    # Set default log level accepted
    _logger.setLevel(logging.INFO)  # Increased to stop matplotlib
    log('\n<<<<<<<<<<<<<<<<<<<<<< LOG SETUP >>>>>>>>>>>>>>>>>>>>>>')


def set_log_level(level):
    global _logger
    _logger.setLevel(level)


def _check_error():
    """
    See if any errors occurred and add warning to end of output
    NB: Expected to be called at or near end of execution to give user easy
    location to if an error occurred
    :return:
    """
    if _ErrorsStats.has_occurred():
        log(f'!!! {_ErrorsStats.COUNT} ERRORS OCCURRED !!!', logging.ERROR)
        _ErrorsStats.notify_err()


class _ErrorsStats:
    COUNT = 0  # Tracks how many errors occurred

    @classmethod
    def register_error_occurrence(cls):
        # add new logger if first error
        if not cls.has_occurred():
            _register_special_error_logger()
        cls.COUNT += 1

    @classmethod
    def has_occurred(cls):
        return cls.COUNT > 0

    @classmethod
    def notify_err(cls):
        from opylib import notify
        notify.notify_err()

    @classmethod
    def register_notification_callback(cls, callback):
        cls.notify_err = callback


def _register_special_error_logger():
    global _logger
    error_handler = logging.FileHandler(os.path.join(_log_dir, 'ERRORS.log'))
    error_handler.setFormatter(
        logging.Formatter('%(asctime)s %(message)s'))
    error_handler.setLevel(logging.ERROR)
    _logger.addHandler(error_handler)
